module Santalo

# This code accompanies the paper "Santalo geometry of convex polytopes" by Dmitrii Pavlov and Simon Telen. 

# Authors: Dmitrii Pavlov and Simon Telen
# Date: February 16, 2024

using Oscar, HomotopyContinuation, LinearAlgebra, IterTools

export free_representation,
    adjoint_x,
    adjoint_y,
    oscar_to_HC_Q,
    solve_likelihood_startsystem,
    get_santalo_point,
    vertex_facet_incidence,
    ideal_XC, 
    compute_Q,
    santalo_path,
    degree_irred_XCuo

# Compute a matrix A and a vector b so that Q can be represented as {x ≥ 0, Ax = b}.
# Input: a full-dimensional polytope Q 
# Output: A, b
function free_representation(Q)
    F = collect(facets(Q))
    W = vcat([-f.a for f in F]...)
    #W = transpose(W)
    c = vcat([f.b for f in F]...)
    A = transpose(nullspace(transpose(W))[2])
    b = A*c
    return [Rational{BigInt}(a) for a in A],  [Rational{BigInt}(bb) for bb in b], [Rational{BigInt}(w) for w in W], [Rational{BigInt}(cc) for cc in c]
end

# Compute the adjoint polynomial in x-coordinates for the cell of the chamber complex containing b in its interior. 
# Input: A and b such that b ∈ pos(A) and {x ≥ 0, Ax = b} is simple 
# Output: A polynomial ring R and the adjoint polynomial adj ∈ R
function adjoint_x(A,b)
    d,n = size(A)
    varstring = ["x$i" for i = 1:n]
    R, vrs = PolynomialRing(QQ,varstring)
    x = vrs
    adj = BigInt(0)
    delem_subsets = vertex_facet_incidence(A,b)
    for ii in delem_subsets
        AA = A[:,setdiff(1:n,ii)]
        adj += abs(Rational{BigInt}(det(AA)))*prod(x[setdiff(1:n,ii)])
    end
    return R, adj
end

# Compute the adjoint polynomial in y-coordinates for the simple polytope Q.
# Input: Q 
# Output: A polynomial ring R and the adjoint polynomial adj ∈ R
function adjoint_y(Q)
    A, b, W, c = free_representation(Q)
    d,n = size(A)
    varstring = ["x$i" for i = 1:n]
    T, vrs = PolynomialRing(QQ,varstring)
    x = vrs
    adj = BigInt(0)
    delem_subsets = vertex_facet_incidence(A,b)
    for ii in delem_subsets
        AA = A[:,setdiff(1:n,ii)]
        adj += abs(det(AA))*prod(x[setdiff(1:n,ii)])
    end
    R, y = PolynomialRing(QQ,["y$j" for j = 1:n-d])
    return R, Oscar.evaluate(adj,W*y+c)
end


# Convert a polynomial f in Oscar format to HomotopyContinuation format. 
# Input: polynomial f to be converted and HomotopyContinuation variables for the result. 
# Optional input: set scale = true if the resulting polynomial should be scaled to make the biggest coefficient 1. 
# Output: the polynomial f in HomotopyContinuation format. 
function oscar_to_HC_Q(f,vars; scale=false, myType = Rational{BigInt})
    if myType == Float64
        cffs = Float64.(convert.(Rational{BigInt},collect(Oscar.coefficients(f))))
    else
        cffs = convert.(myType,collect(Oscar.coefficients(f)))
    end
    exps = collect(Oscar.exponents(f))
    if scale
        cffs = cffs/maximum(abs.(cffs))
    end
    sum([cffs[i]*prod(vars.^exps[i]) for i = 1:length(cffs)])
end

# Solve the likelihood equations for the Wachspress model of a simple polytope {x ≥ 0, Ax = myb}, for generic data u₀, and righthand side myb.
# Input: the matrix A and righthand side myb
# Output: the parametric sytem of equations S representing F(x;u) = 0 and the result monres of the command monodromy_solve(S)
#       solutions(monres) contains all complex critical points
#       length(solutions(monres)) is the ML degree of the Wachspress model
#       parameters(monres) is the vector u₀
function solve_likelihood_startsystem(A,myb)
    R, adj = adjoint_x(A,myb)
    d,n = size(A)
    @var x[1:n] 
    @var u[1:n+1]
    adj_HC = oscar_to_HC_Q(adj,x;scale = true)

    B = transpose(nullspace(Float64.(A)))
    eqs = B*[u[n+1]*differentiate(adj_HC,x[i])/adj_HC - u[i]/x[i] for i = 1:n]
    lineqs = A*x-myb

    x₀ = pinv(Float64.(A))*Float64.(myb)
    lineqsu = HomotopyContinuation.expand.(System(eqs,variables = x, parameters = u)(x₀))
    N = nullspace(transpose(hcat([exponents_coefficients(l+eps()*sum(u),u)[2] for l in lineqsu]...)))
    u₀ = N*randn(ComplexF64,size(N,2))
    S = System([eqs;lineqs], variables = x, parameters = u) 
    monres = monodromy_solve(S,[x₀],u₀; compile = false)
    return S, monres
end

# Check numerically whether the variety \mathscr{X}_{C,myu}^\circ is irreducible, and compute the degree of its closure. 
# Input: the matrix A, a righthand side myb in the interior of the cell C, and a data vector myu 
# Output: a boolean irred indicating whether the variety is irreducible or not, and an integer deg representing its degree 
function degree_irred_XCuo(A,myb,myu)
    R, adj = adjoint_x(A,myb)
    d,n = size(A)
    @var x[1:n] 
    @var u[1:n+1]
    @var AA[1:d,1:n]
    @var bb[1:d]
    adj_HC = oscar_to_HC_Q(adj,x;scale = true)

    B = transpose(nullspace(Float64.(A)))
    eqs = B*[u[n+1]*differentiate(adj_HC,x[i])/adj_HC - u[i]/x[i] for i = 1:n]
    lineqs = AA*x-bb

    AA0 = randn(size(A))
    b0 = randn(length(myb))
    x₀ = pinv(Float64.(AA0))*Float64.(b0)
    lineqsu = HomotopyContinuation.expand.(System(eqs,variables = x, parameters = u)(x₀))
    N = nullspace(transpose(hcat([exponents_coefficients(l+eps()*sum(u),u)[2] for l in lineqsu]...)))
    u₀ = N*randn(ComplexF64,size(N,2))
    S = System([eqs;lineqs], variables = x, parameters = [u;AA[:];bb]) 
    monres = monodromy_solve(S,[x₀],[u₀;AA0[:];b0]; compile = false)
    R = HomotopyContinuation.solve(S, solutions(monres); start_parameters = parameters(monres), target_parameters = [myu;AA0[:];b0])
    deg = length(solutions(R))
    T = System([subs(eqs,u=>myu);lineqs],variables = x, parameters = [AA[:];bb])
    newmonres = monodromy_solve(T,[solutions(R)[1]],[AA0[:];b0])
    irred = false
    if length(solutions(newmonres)) == deg
        irred = true
    end
    return irred, deg
end


# Compute the Santalo point y^* of a full dimensional simple polytope Q.
# Input: a polytope Q
# Output: y^*
function get_santalo_point(Q)
    A, myb, W, c = free_representation(Q)
    d, n = size(A)
    S, monres = solve_likelihood_startsystem(A,myb)
    R = HomotopyContinuation.solve(S,solutions(monres); start_parameters=parameters(monres), target_parameters = ones(n+1))
    println("$(length(findall(s->prod(s.>0) == 1, HomotopyContinuation.real_solutions(R)))) positive critical point(s)")
    santalo_index = findfirst(s->prod(s.>0) == 1, HomotopyContinuation.real_solutions(R))
    return pinv(Float64.(W))*(HomotopyContinuation.real_solutions(R)[santalo_index]-Float64.(c))
end


# Compute the Santalo point x^* of a full dimensional simple polytope {x ≥ 0, Ax = myb}.
# Input: A, myb
# Output: x^*
function get_santalo_point(A, myb)
    d, n = size(A)
    S, monres = solve_likelihood_startsystem(A,myb)
    R = HomotopyContinuation.solve(S,solutions(monres); start_parameters=parameters(monres), target_parameters = ones(n+1))
    santalo_index = findfirst(s->prod(s.>0) == 1, HomotopyContinuation.real_solutions(R))
    return HomotopyContinuation.real_solutions(R)[santalo_index]
end

# For each vertex of {x ≥ 0, Ax = myb}, compute which facets meet at that vertex.
# Input: A and myb
# Output: a vector containing, for each vertex, a list of facet indices which meat at the vertex.
function vertex_facet_incidence(A,myb)
    d,n = size(A)
    B = nullspace(MatrixSpace(QQ,size(A)...)(A))[2]
    B = [Rational{BigInt}(b) for b in B]
    T = [A; transpose(B)]

    MS = MatrixSpace(QQ,size(T)...)
    invT = inv(MS(T))
    C = invT[:,1:d]
    W = invT[:,d+1:n]

    Q = Polyhedron(-W,C*myb)
    VQ = collect(vertices(Q))

    VQ =[convert(Vector{Rational{BigInt}},v) for v in VQ]
    [findall(i-> (C[i,:]*myb+W[i,:]*v)[1] == 0, 1:size(C,1)) for v in VQ]
end

# Compute the ideal of the variety {\cal X}_C, where C is the cell of myb in the chamber complex of A. 
# Input: A, myb
# Output: a polynomial ring containing the ideal of {\cal X}_C, and the ideal itself 
function ideal_XC(A,myb)
    d,n = size(A)
    R, adj = adjoint_x(A,myb)
    B = nullspace(MatrixSpace(QQ,size(A)...)(A))[2]
    B = [Rational{BigInt}(b) for b in B]
    FC = union(vertex_facet_incidence(A,myb)...)
    xstring = ["x$i" for i = 1:n]
    sigmastring = ["σ$i" for i = 1:n]
    S,vrs  = PolynomialRing(QQ,[xstring;sigmastring;"z"])
    x = vrs[1:n]
    σ = vrs[n+1:end-1]
    z = vrs[end]
    adj = Oscar.evaluate(adj,x)
    eqs = transpose(B)*[z*derivative(adj,x[i]) - σ[i]*(i in FC) for i in 1:n]
    J = ideal(S,[eqs;[σ[i]*x[i]-1 for i = 1:n]; z*adj-1])
    E = eliminate(J,[σ;z])
    T, x = graded_polynomial_ring(QQ,xstring)
    Jnew = [Oscar.evaluate(g,[x;zeros(T,length(x)+1)]) for g in gens(E)]
    return T, ideal(T,Jnew)
end

# Compute the projected polytope Q = B^T ̇Pb, where Pb = {x ≥ 0, Ax = b} and B = ker A. 
# Input: A, b and B
# Output: the projected polytope Q
function compute_Q(A,b,B)
    d,n = size(A)
    T = [A;transpose(B)]
    invT = inv(T)
    c = invT[:,1:d]*b
    W = invT[:,d+1:end]
    polyhedron(-W,c)
end

# Turn a matrix into LaTeX format
function to_tex(M)
    Mstring = "\\begin{matrix} \n "
    for i = 1:size(M,1)
        rowstring = prod([string(M[i,j])*" & " for j = 1:size(M,2)])
        Mstring = Mstring*rowstring[1:end-2]*" \\\\ \n "
    end
    Mstring = Mstring*"\\end{matrix}"
end

# Compute the Santalo point x^*(b1) from x^*(b0), where b0 and b1 belong to the same cell of the chamber complex of A. 
# Input: A, b0, b1, and xstar0 = x^*(b0)
# Output: x^*(b1)
function santalo_path(A,b0,b1,xstar0)
    R, adj = adjoint_x(A,b0)
    d,n = size(A)
    @var x[1:n] 
    @var b[1:d]
    adj_HC = oscar_to_HC_Q(adj,x;scale = true)

    B = transpose(nullspace(Float64.(A)))
    eqs = B*[differentiate(adj_HC,x[i])/adj_HC - 1/x[i] for i = 1:n]
    lineqs = A*x-b

    S = System([eqs;lineqs], variables = x, parameters = b) 
    println(S(xstar0,b0))
    R = HomotopyContinuation.solve(S,[xstar0]; start_parameters = b0, target_parameters = b1)
    println(HomotopyContinuation.real_solutions(R))
    if isempty(HomotopyContinuation.real_solutions(R))
        println("path tracking failed")
        return []
    else
        return HomotopyContinuation.real_solutions(R)[1]
    end
end


end # module Santalo
